theory Ordered_Algebra
  imports "$AFP/Automatic_Refinement/Lib/Misc"
   "Quasi_Order" "~~/HOL/Lattices"
begin

section {* F-Algebras *}

locale algebra = fixes I :: "'f \<Rightarrow> 'a list \<Rightarrow> 'a"
begin

fun eval :: "('f,'v) term \<Rightarrow> ('v \<Rightarrow> 'a) \<Rightarrow> 'a"
  where "eval (Var x) \<alpha> = \<alpha> x"
      | "eval (Fun f ss) \<alpha> = I f (map (\<lambda>s. eval s \<alpha>) ss)"

notation eval ("\<lbrakk>(_)\<rbrakk>")

definition subst_assign where "subst_assign \<sigma> \<alpha> x \<equiv> \<lbrakk>\<sigma> x\<rbrakk>\<alpha>"

lemma subst_assign: "\<lbrakk>s\<rbrakk>(subst_assign \<sigma> \<alpha>) = \<lbrakk>s\<cdot>\<sigma>\<rbrakk>\<alpha>"
proof(induct s)
  case (Var x) show ?case unfolding subst_assign_def by auto
next
  case (Fun f ss)
  { fix x note impI[of "x \<in> set ss", OF Fun[of x], simplified] }
  note * = map_ext[OF this]
  show ?case by (auto simp: comp_def *)
qed

lemma eval_same_vars[intro]: "(\<And>x. x \<in> vars_term t \<Longrightarrow> \<alpha> x = \<beta> x) \<Longrightarrow> \<lbrakk>t\<rbrakk> \<alpha> = \<lbrakk>t\<rbrakk> \<beta>"
  by(induct t, auto intro!:map_cong[OF refl] cong[of "I _"])

lemma subset_range_eval: "range \<alpha> \<subseteq> range (\<lambda>s. \<lbrakk>s\<rbrakk>\<alpha>)" (is "?l \<subseteq> ?r")
proof
  fix a assume "a \<in> ?l"
  then obtain v where "\<alpha> v = a" by auto
  then show "a \<in> ?r" by (auto intro!: range_eqI[of _ _ "Var v"])
qed


end

lemma eval_map_term[simp]:
  "algebra.eval I (map_term ff fv t) \<alpha> = algebra.eval (I \<circ> ff) t (\<alpha> \<circ> fv)"
  by (induct t, auto simp: o_def algebra.eval.simps intro: cong[of "I _"])

locale term_algebra = algebra "Fun :: 'f \<Rightarrow> ('f,'v) term list \<Rightarrow> ('f,'v) term"
begin

lemma interpret_is_subst[simp]: "\<lbrakk>s\<rbrakk>\<sigma> = s\<cdot>\<sigma>" by (induct s, unfold algebra.eval.simps, auto)

lemma subst_assign_is_comp[simp]: "subst_assign \<sigma> \<tau> = \<sigma> \<circ>\<^sub>s \<tau>" by (auto simp: algebra.subst_assign_def subst_compose)

end

subsection {* Quasi-ordered algebra *}

text
{* A quasi-ordered algebra is a F-algebra whose domain is quasi-ordered.
   We cannot use @{class quasi_order} as a class since later we want to use
   arbitrary order on terms which cannot be globally fixed. Local instantiation
   would alleviate the problem.
 *}

locale ordered_algebra = quasi_order less_eq less + algebra I
  for less_eq less :: "'a \<Rightarrow> 'a \<Rightarrow> bool"
  and I :: "'f \<Rightarrow> 'a list \<Rightarrow> 'a"
begin

interpretation ord_syntax less_eq less.

text{* Base orderings are extended over terms in the standard way. *}

definition less_term (infix "\<sqsubset>\<^sub>t" 50) where "s \<sqsubset>\<^sub>t t \<equiv> \<forall>\<alpha>. \<lbrakk>s\<rbrakk>\<alpha> \<sqsubset> \<lbrakk>t\<rbrakk>\<alpha>"
notation less_term (infix "\<sqsubset>\<^sub>t" 50)

definition less_eq_term (infix "\<sqsubseteq>\<^sub>t" 50) where "s \<sqsubseteq>\<^sub>t t \<equiv> \<forall>\<alpha>. \<lbrakk>s\<rbrakk>\<alpha> \<sqsubseteq> \<lbrakk>t\<rbrakk>\<alpha>"
notation less_eq_term (infix "\<sqsubseteq>\<^sub>t" 50)

lemma less_termI[intro]: "(\<And>\<alpha> :: 'v \<Rightarrow> 'a. \<lbrakk>s\<rbrakk>\<alpha> \<sqsubset> \<lbrakk>t\<rbrakk>\<alpha>) \<Longrightarrow> s \<sqsubset>\<^sub>t t" by (auto simp: less_term_def)
lemma less_termD[dest]: fixes \<alpha> :: "'v \<Rightarrow> 'a" shows "s \<sqsubset>\<^sub>t t \<Longrightarrow> \<lbrakk>s\<rbrakk>\<alpha> \<sqsubset> \<lbrakk>t\<rbrakk>\<alpha>" by (auto simp: less_term_def)

lemma less_eq_termI[intro]: "(\<And>\<alpha> :: 'v \<Rightarrow> 'a. \<lbrakk>s\<rbrakk>\<alpha> \<sqsubseteq> \<lbrakk>t\<rbrakk>\<alpha>) \<Longrightarrow> s \<sqsubseteq>\<^sub>t t" by (auto simp: less_eq_term_def)
lemma less_eq_termD[dest]: fixes \<alpha> :: "'v \<Rightarrow> 'a" shows "s \<sqsubseteq>\<^sub>t t \<Longrightarrow> \<lbrakk>s\<rbrakk>\<alpha> \<sqsubseteq> \<lbrakk>t\<rbrakk>\<alpha>" by (auto simp: less_eq_term_def)

sublocale "term": quasi_order "op \<sqsubseteq>\<^sub>t" "op \<sqsubset>\<^sub>t"
  apply (unfold_locales, unfold less_term_def less_eq_term_def)
  apply (auto dest: order_trans less_trans le_less_trans less_le_trans less_imp_le)
  done

notation(input) term.greater (infix "\<sqsupset>\<^sub>t" 50)
notation(input) term.greater_eq (infix "\<sqsupseteq>\<^sub>t" 50)

notation term.less_eq_set ("{\<sqsubseteq>\<^sub>t}")
notation term.less_set ("{\<sqsubset>\<^sub>t}")
notation term.greater_eq_set ("{\<sqsupseteq>\<^sub>t}")
notation term.greater_set ("{\<sqsupset>\<^sub>t}")

notation term.equiv (infix "\<simeq>\<^sub>t" 50)
notation term.equiv_class ("[_]\<^sub>t")

lemma equiv_termI[intro]: "(\<And>\<alpha> :: 'v \<Rightarrow> 'a. \<lbrakk>s\<rbrakk>\<alpha> \<simeq> \<lbrakk>t\<rbrakk>\<alpha>) \<Longrightarrow> s \<simeq>\<^sub>t t" by auto
lemma equiv_termD[dest]: fixes \<alpha> :: "'v \<Rightarrow> 'a" shows "s \<simeq>\<^sub>t t \<Longrightarrow> \<lbrakk>s\<rbrakk>\<alpha> \<simeq> \<lbrakk>t\<rbrakk>\<alpha>" by auto

lemma equiv_term: fixes s t :: "('f,'v) term" shows "s \<simeq>\<^sub>t t \<longleftrightarrow> s \<sqsubseteq>\<^sub>t t \<and> t \<sqsubseteq>\<^sub>t s" by fastforce

lemma less_term_stable[intro]: fixes \<sigma> :: "'v \<Rightarrow> ('f,'w) term" shows "s \<sqsubset>\<^sub>t t \<Longrightarrow> s\<cdot>\<sigma> \<sqsubset>\<^sub>t t\<cdot>\<sigma>"
  by (intro less_termI, fold subst_assign, auto)
lemma less_eq_term_stable[intro]: fixes \<sigma> :: "'v \<Rightarrow> ('f,'w) term" shows "s \<sqsubseteq>\<^sub>t t \<Longrightarrow> s\<cdot>\<sigma> \<sqsubseteq>\<^sub>t t\<cdot>\<sigma>"
  by (intro less_eq_termI, fold subst_assign, auto)

end

locale wf_algebra = ordered_algebra less_eq less I + wf_order less_eq less
  for less_eq less :: "'a \<Rightarrow> 'a \<Rightarrow> bool" and I :: "'f \<Rightarrow> 'a list \<Rightarrow> 'a"
begin

interpretation ord_syntax less_eq less.

sublocale "term": wf_order "op \<sqsubseteq>\<^sub>t :: ('f,'v) term \<Rightarrow> ('f,'v) term \<Rightarrow> bool" "op \<sqsubset>\<^sub>t"
proof
  obtain v where "(v :: 'a) = v" by auto
  define \<alpha> where "\<alpha> \<equiv> \<lambda>x :: 'v. v"
  fix P s
  assume *: "\<And>s :: ('f,'v) term. (\<And>t. t \<sqsubset>\<^sub>t s \<Longrightarrow> P t) \<Longrightarrow> P s"
  then show "P s"
  proof(induct "\<lbrakk>s\<rbrakk>\<alpha>" arbitrary: s rule: less_induct)
    case less show ?case
    proof (rule less(2))
      fix t assume ts: "t \<sqsubset>\<^sub>t s"
      then have "\<lbrakk>t\<rbrakk>\<alpha> \<sqsubset> \<lbrakk>s\<rbrakk>\<alpha>" by auto
      from less(1)[OF this *] show "P t" by auto
    qed
  qed
qed

end


lemma(in quasi_order) weak_mono_iff_all:
  "(\<forall>a b ls rs. a \<le> b \<longrightarrow> f (ls @ a # rs) \<le> f (ls @ b # rs)) \<longleftrightarrow>
   (\<forall>as bs. length as = length bs \<longrightarrow> (\<forall>i < length as. as ! i \<le> bs ! i) \<longrightarrow> f as \<le> f bs)"
  (is "?l \<longleftrightarrow> ?r")
proof(intro iffI impI allI)
  fix as bs
  assume mono: ?l and all: "\<forall>i < length as. as ! i \<le> bs ! i" and len: "length as = length bs"
  { fix cs
    from all have "\<And>i. i < length as \<Longrightarrow> as ! i \<le> bs ! i" by auto
    with len have "f (cs @ as) \<le> f (cs @ bs)"
    proof(induct as arbitrary: bs cs)
      case Nil then have "bs = []" by auto
      with order_refl show ?case by auto
    next
      case IH: (Cons a as)
      then have "Suc (length as) = length bs" by auto
      then obtain d ds where bs: "bs = d # ds" and len: "length as = length ds" by (induct bs, auto)
      have "f ((cs @ [a]) @ as) \<le> f ((cs @ [a]) @ ds)"
        apply(rule IH(1)[OF len]) using IH(3)[unfolded bs] by force
      also have "... \<le> f ((cs @ [d]) @ ds)"
        proof-
          from IH(3)[unfolded bs, of 0] have "a \<le> d" by auto
          with mono show ?thesis by auto
        qed
      finally show ?case by (unfold bs, auto)
    qed
  }
  from this[of "[]"] show "f as \<le> f bs" by auto
next
  assume ?r
  then have *: "\<And>as bs. length as = length bs \<Longrightarrow> (\<forall>i<length as. as ! i \<le> bs ! i) \<Longrightarrow> f as \<le> f bs"
    by auto
  fix a b ls rs
  assume ab: "a \<le> b"
  show "f (ls @ a # rs) \<le> f (ls @ b # rs)"
  proof (rule *,simp, intro allI impI)
    fix i assume i: "i < length (ls @ a # rs)"
    show " (ls @ a # rs) ! i \<le> (ls @ b # rs) ! i"
    by(cases i "length ls" rule: linorder_cases, auto simp: ab nth_append)
  qed
qed

locale weak_mono_fun = quasi_order +
  fixes f
  assumes weak_mono: "\<And>ls rs a b. less_eq a b \<Longrightarrow> less_eq (f (ls @ a # rs)) (f (ls @ b # rs))"
begin

lemma weak_mono_all:
  assumes "length as = length bs"
    and "\<And>i. i < length as \<Longrightarrow> less_eq (as ! i) (bs ! i)"
  shows "less_eq (f as) (f bs)"
proof-
  from weak_mono
  have "\<forall>as bs. length as = length bs \<longrightarrow> (\<forall>i<length as. less_eq (as ! i) (bs ! i)) \<longrightarrow> less_eq (f as) (f bs)"
    by(subst weak_mono_iff_all[symmetric], auto)
  with assms show ?thesis by auto
qed

end

subsection {* Weakly monotone algebra *}

locale weak_mono_algebra = ordered_algebra +
  assumes weak_mono: "\<And>f ls rs a b. less_eq a b \<Longrightarrow> less_eq (I f (ls @ a # rs)) (I f (ls @ b # rs))"
begin

sublocale "fun": weak_mono_fun less_eq less "I f" by(unfold_locales, fact weak_mono)

sublocale "term": weak_mono_fun less_eq_term less_term "Fun f"
  by (unfold_locales, intro less_eq_termI, auto intro!: weak_mono)

interpretation ord_syntax less_eq less.

lemma ctxt_closed_NS: "ctxt.closed {\<sqsupseteq>\<^sub>t}"
 by (rule one_imp_ctxt_closed, auto intro: term.weak_mono)

lemma subst_closed_S: "subst.closed {\<sqsupset>\<^sub>t}" by auto

lemma subst_closed_NS: "subst.closed {\<sqsupseteq>\<^sub>t}" by auto

lemma le_ctxt_closed[intro]: "s \<sqsupseteq>\<^sub>t t \<Longrightarrow> C\<langle>s\<rangle> \<sqsupseteq>\<^sub>t C\<langle>t\<rangle>"
  using ctxt.closedD[OF ctxt_closed_NS] by auto

lemma mono_range: "(\<And>v. v \<in> vars_term s \<Longrightarrow> \<alpha> v \<sqsubseteq> \<beta> v) \<Longrightarrow> \<lbrakk>s\<rbrakk>\<alpha> \<sqsubseteq> \<lbrakk>s\<rbrakk>\<beta>"
proof (induct s)
  case (Var v) then show ?case by auto
next
  case IH: (Fun f ss)
  { fix i assume i: "i < length ss"
    then have "vars_term (ss!i) \<subseteq> vars_term (Fun f ss)" by fastforce
    with IH(2) have *: "v \<in> vars_term (ss!i) \<Longrightarrow> \<alpha> v \<sqsubseteq> \<beta> v" for v by blast
    with i have "\<lbrakk>ss ! i\<rbrakk>\<alpha> \<sqsubseteq> \<lbrakk>ss ! i\<rbrakk>\<beta>" by(intro IH(1), auto)
  }
  then show ?case by(auto intro: fun.weak_mono_all)
qed

end

locale wf_weak_mono_algebra = weak_mono_algebra + wf_order
begin

sublocale wf_algebra..

sublocale redpair: redpair "{\<sqsupset>\<^sub>t}" "{\<sqsupseteq>\<^sub>t}"
  using SN ctxt_closed_NS subst_closed_S subst_closed_NS by unfold_locales

end


locale mono_algebra = weak_mono_algebra +
  assumes mono: "\<And>f ls rs a b. less a b \<Longrightarrow> less (I f (ls @ a # rs)) (I f (ls @ b # rs))"
begin

lemma mono_term: "s \<sqsubset>\<^sub>t t \<Longrightarrow> Fun f (ls @ s # rs) \<sqsubset>\<^sub>t Fun f (ls @ t # rs)"
  by(intro less_termI, auto intro: mono)

lemma ctxt_closed_S: "ctxt.closed {\<sqsupset>\<^sub>t}"
  by (rule one_imp_ctxt_closed, auto intro: mono_term)

lemma less_ctxt_closed[intro]: "s \<sqsupset>\<^sub>t t \<Longrightarrow> C\<langle>s\<rangle> \<sqsupset>\<^sub>t C\<langle>t\<rangle>"
  using ctxt.closedD[OF ctxt_closed_S] by auto

end

locale wf_mono_algebra = mono_algebra + wf_order
begin
  sublocale wf_weak_mono_algebra..
end

section {* Models *}

locale quasi_model = weak_mono_algebra less_eq less I
  for less_eq less and I :: "'f \<Rightarrow> 'a list \<Rightarrow> 'a" +
  fixes R :: "('f,'v) trs"
  assumes R_imp_le: "(l,r) \<in> R \<Longrightarrow> l \<sqsupseteq>\<^sub>t r"
begin

lemma rsteps_imp_ge: "(s,t) \<in> (rstep R)\<^sup>* \<Longrightarrow> s \<sqsupseteq>\<^sub>t t"
proof(induct rule: rtrancl_induct)
  case base show ?case by auto
next
  case (step t u)
  then obtain C \<sigma> l r where t: "t = C\<langle>l\<cdot>\<sigma>\<rangle>" and u: "u = C\<langle>r\<cdot>\<sigma>\<rangle>" and lr: "(l,r) \<in> R" by auto
  have "t \<sqsupseteq>\<^sub>t u" by (unfold t u, auto intro: R_imp_le lr)
  also note step(3)
  finally show ?case.
qed

end

locale strict_model = mono_algebra less_eq less I
  for less_eq less and I :: "'f \<Rightarrow> 'a list \<Rightarrow> 'a" +
  fixes R :: "('f,'v) trs"
  assumes R_imp_less: "(l,r) \<in> R \<Longrightarrow> l \<sqsupset>\<^sub>t r"
begin

sublocale quasi_model using R_imp_less term.less_imp_le by (unfold_locales, auto)

lemma rstep_imp_greater: assumes st: "(s,t) \<in> rstep R" shows "s \<sqsupset>\<^sub>t t"
proof-
  from st obtain C \<sigma> l r where s: "s = C\<langle>l\<cdot>\<sigma>\<rangle>" and t: "t = C\<langle>r\<cdot>\<sigma>\<rangle>" and lr: "(l,r) \<in> R" by auto
  show "s \<sqsupset>\<^sub>t t" by (unfold s t, auto intro: R_imp_less lr)
qed

lemma rsteps_imp_greater: "(s,t) \<in> (rstep R)\<^sup>+ \<Longrightarrow> s \<sqsupset>\<^sub>t t"
  by (induct rule: trancl_induct, auto dest:rstep_imp_greater term.less_trans)

end

locale wf_model = strict_model + wf_order
begin

sublocale wf_mono_algebra..

lemma SN_rstep: "SN (rstep R)"
  by (unfold SN_iff_wf, rule wf_subset, rule term.wf, auto intro: rstep_imp_greater)

end

text {* A TRS induces a monotone algebra w.r.t. rewriting. *}

locale rewrite_algebra = fixes R :: "('f,'v) trs"
begin

definition less_eq (infix "\<sqsubseteq>" 50) where "less_eq s t \<equiv> (t,s) \<in> (rstep R)\<^sup>*"
definition less (infix "\<sqsubset>" 50) where "less s t \<equiv> (t,s) \<in> (rstep R)\<^sup>+"

sublocale term_algebra.

sublocale mono_algebra less_eq less Fun
  using ctxt_closed_rsteps[of R] ctxt.closed_trancl[OF ctxt_closed_rstep[of R]]
  by (unfold_locales, auto simp: less_eq_def less_def dest: ctxt_closed_one)

lemma less_term_is_less[simp]: "less_term = less"
  using subst.closed_trancl[OF subst_closed_rstep[of R]]
  by (unfold less_term_def, auto intro!: ext elim: allE[of _ Var] simp: less_def)

lemma le_term_is_le[simp]: "less_eq_term = less_eq"
  using subst.closed_rtrancl[OF subst_closed_rstep[of R]]
  by (unfold less_eq_term_def, auto intro!: ext elim: allE[of _ Var] simp: less_eq_def)

lemma less_stable: "s \<sqsubset> t \<Longrightarrow> s\<cdot>\<sigma> \<sqsubset> t\<cdot>\<sigma>"
  using less_term_stable[where 'v='v and 'w='v] by simp

lemma less_eq_stable: "s \<sqsubseteq> t \<Longrightarrow> s\<cdot>\<sigma> \<sqsubseteq> t\<cdot>\<sigma>"
  using less_eq_term_stable[where 'v='v and 'w='v] by simp

sublocale strict_model less_eq less Fun by (unfold_locales, auto simp: less_def)

end

text {* A TRS is terminating if and only if there is a well-founded model for it.
  In paper we don't restrict the domain of interpretations but in Isabelle/HOL we have to.
*}
proposition ex_wf_model_iff_terminating:
  fixes R :: "('f,'v) trs"
  shows "(\<exists>less_eq less :: ('f,'v) term \<Rightarrow> _. \<exists> I. wf_model less_eq less I R) \<longleftrightarrow> SN (rstep R)" (is "?l \<longleftrightarrow> ?r")
proof
  assume ?l
  then obtain less_eq less :: "('f,'v) term \<Rightarrow> _" and I where "wf_model less_eq less I R" by auto
  then interpret wf_model less_eq less I R by auto
  from SN_rstep show ?r.
next
  interpret rewrite_algebra R.
  assume ?r
  then have "SN ((rstep R)\<^sup>+)" by (rule SN_on_trancl)
  from SN_induct[OF this]
  have "wf_model less_eq less Fun R" by (unfold_locales, auto simp: less_def)
  then show ?l by auto
qed

locale equimodel = weak_mono_algebra less_eq less I
  for less_eq less and I :: "'f \<Rightarrow> 'a list \<Rightarrow> 'a" +
  fixes E :: "('f,'v) trs"
  assumes E_imp_equiv: "(l,r) \<in> E \<Longrightarrow> l \<simeq>\<^sub>t r"
begin

sublocale quasi_model less_eq less I "E\<^sup>\<leftrightarrow>"
  by (unfold_locales, insert E_imp_equiv, auto)

lemma conversion_imp_equiv: assumes "(s,t) \<in> (rstep E)\<^sup>\<leftrightarrow>\<^sup>*" shows "s \<simeq>\<^sub>t t"
proof-
  note * = rsteps_imp_ge[unfolded rstep_simps(5), folded conversion_def]
  from *[OF assms[unfolded conversion_inv[of s t]]] *[OF assms]
  show ?thesis by (auto simp: equiv_term)
qed

end

locale model = algebra I for I :: "'f \<Rightarrow> 'a list \<Rightarrow> 'a" and E :: "('f,'v) trs" +
  assumes E_imp_eq: "(l,r) \<in> E \<Longrightarrow> \<lbrakk>l\<rbrakk> = \<lbrakk>r\<rbrakk>"
begin

interpretation ordered: ordered_algebra "op =" "\<lambda>x y. False" I by (unfold_locales, auto)
sublocale ordered: equimodel "op =" "\<lambda>x y. False" I E by (unfold_locales, auto dest!: E_imp_eq)

lemma conversion_imp_eq: assumes "(s,t) \<in> (rstep E)\<^sup>\<leftrightarrow>\<^sup>*" shows "\<lbrakk>s\<rbrakk> = \<lbrakk>t\<rbrakk>"
  using ordered.conversion_imp_equiv[OF assms] by auto

end

subsection {* Homomorphisms *}

locale algebra_hom = algebra I + target: algebra I'
  for hom :: "'a \<Rightarrow> 'b" and I :: "'f \<Rightarrow> 'a list \<Rightarrow> 'a" and I' :: "'f \<Rightarrow> 'b list \<Rightarrow> 'b" +
  assumes hom: "\<And>f as. hom (I f as) = I' f (map hom as)"
begin

notation target.eval ("\<lbrakk>(_)\<rbrakk>''")

lemma hom_eval: "hom (\<lbrakk>s\<rbrakk>\<alpha>) = \<lbrakk>s\<rbrakk>'(hom \<circ> \<alpha>)"
  by (induct s, auto simp add: hom o_def intro: arg_cong[OF list.map_cong0])

end

locale ordered_algebra_hom =
  algebra_hom hom I I' + ordered_algebra less_eq less I +
  target: ordered_algebra less_eq' less' I'
  for hom less_eq less I less_eq' less' I' +
  assumes le_hom: "less_eq a b \<Longrightarrow> less_eq' (hom a) (hom b)"
      and less_hom: "less a b \<Longrightarrow> less' (hom a) (hom b)"

locale algebra_epim = algebra_hom +
  assumes surj: "surj hom"

locale ordered_algebra_epim = ordered_algebra_hom + algebra_epim

definition models_less
  where "models_less (ty :: 'a itself) R t s \<longleftrightarrow>
  (\<forall>less_eq less :: 'a \<Rightarrow> _. \<forall> I. strict_model less_eq less I R \<longrightarrow> ordered_algebra.less_term less I t s)"

context strict_model begin
  interpretation ord_syntax.
  lemma models_lessD:
    assumes "models_less (TYPE('a)) R t s" shows "t \<sqsubset>\<^sub>t s"
    using assms strict_model_axioms by (auto simp: models_less_def)
end

lemma models_less_imp_rsteps:
  assumes "models_less (TYPE(('f,'v) term)) (R::('f,'v) trs) t s"
  shows "(s,t) \<in> (rstep R)\<^sup>+"
proof-
  interpret rewrite_algebra R.
  from models_lessD[OF assms] show ?thesis by (auto simp: less_def)
qed

lemma rsteps_imp_models_less:
  assumes st: "(s,t) \<in> (rstep R)\<^sup>+"
  shows "models_less (TYPE('a)) R t s"
proof (unfold models_less_def, intro allI impI)
  interpret rewrite_algebra R.
  fix less_eq less and I :: "_ \<Rightarrow> 'a list \<Rightarrow> 'a"
  assume "strict_model less_eq less I R"
  then interpret target: strict_model less_eq less I R.
  from target.rsteps_imp_greater[OF st]
  show "target.less_term t s" by auto
qed

context rewrite_algebra begin

  proposition strict_model_imp_hom:
    assumes model: "strict_model (less_eq' :: 'b \<Rightarrow> _) less' I' R"
    shows "\<exists>hom. ordered_algebra_hom hom less_eq less Fun less_eq' less' I'"
  proof-
    interpret target: strict_model less_eq' less' I' R using model.
    note connect = target.rsteps_imp_ge[folded less_eq_def] target.rsteps_imp_greater[folded less_def]
    have "ordered_algebra_hom (\<lambda>s. target.eval s \<alpha>) less_eq less Fun less_eq' less' I'" for \<alpha>
      by (unfold_locales, auto dest: connect)
    then show ?thesis by auto
  qed

  theorem strict_model_imp_epim:
    assumes surj: "surj (\<alpha> :: 'v \<Rightarrow> 'b)" (* there should be enough variables *)
        and model: "strict_model (less_eq' :: 'b \<Rightarrow> _) less' I' R"
    defines "hom \<equiv> \<lambda>s. algebra.eval I' s \<alpha>"
    shows "ordered_algebra_epim hom less_eq less Fun less_eq' less' I'"
  proof-
    interpret target: strict_model less_eq' less' I' R using model.
    note connect = target.rsteps_imp_ge[folded less_eq_def] target.rsteps_imp_greater[folded less_def]
    from surj have "surj hom" using target.subset_range_eval[of \<alpha>] by (auto simp: hom_def)
    then show "ordered_algebra_epim hom less_eq less Fun less_eq' less' I'"
      by (unfold_locales, auto dest: connect simp: hom_def)
  qed

end


subsection {* Encoding algebras *}

text {* Encoding $F$-algebra into $G$-algebra.
 Each function symbol $f \in F$ of arity $n$ is encoded as a term in $T(G,{0..<n})$. *}

locale encoding =
  fixes E :: "'f \<times> nat \<Rightarrow> ('g, nat) term"
  assumes var_domain: "i \<in> vars_term (E (f,n)) \<Longrightarrow> i < n"
begin



end


locale encoded_algebra = target: algebra I + encoding E
  for I :: "'g \<Rightarrow> 'a list \<Rightarrow> 'a"
  and E :: "'f \<times> nat \<Rightarrow> ('g, nat) term"
begin

definition IE :: "'f \<Rightarrow> 'a list \<Rightarrow> 'a" where "IE f vs = target.eval (E (f, length vs)) (nth vs)"

sublocale algebra IE.

end

locale encoded_ordered_algebra =
  quasi_order less_eq less + encoding E
  for less_eq less :: "'a \<Rightarrow> 'a \<Rightarrow> bool"
  and I :: "'g \<Rightarrow> 'a list \<Rightarrow> 'a"
  and E :: "'f \<times> nat \<Rightarrow> ('g, nat) term"
begin

sublocale encoded_algebra I E..

sublocale target: ordered_algebra less_eq less I..

sublocale ordered_algebra less_eq less IE..

end

locale encoded_wf_algebra =
  wf_order less_eq less + encoding E
  for less_eq less :: "'a \<Rightarrow> 'a \<Rightarrow> bool"
  and I :: "'g \<Rightarrow> 'a list \<Rightarrow> 'a"
  and E :: "'f \<times> nat \<Rightarrow> ('g, nat) term"
begin

sublocale encoded_ordered_algebra less_eq less I E..

sublocale target: wf_algebra less_eq less I by (unfold_locales, fact less_induct)

sublocale wf_algebra less_eq less IE..

end


locale encoded_weak_mono_algebra =
  quasi_order less_eq less + encoding E
  for less_eq less :: "'a \<Rightarrow> 'a \<Rightarrow> bool"
  and I :: "'g \<Rightarrow> 'a list \<Rightarrow> 'a"
  and E :: "'f \<times> nat \<Rightarrow> ('g, nat) term" +
  assumes mono: "\<And>f ls rs a b. less_eq a b \<Longrightarrow> less_eq (I f (ls @ a # rs)) (I f (ls @ b # rs))"

begin

sublocale target: weak_mono_algebra less_eq less I by (unfold_locales, fact mono)

sublocale encoded_ordered_algebra less_eq less I E..

interpretation ord_syntax less_eq less.

sublocale weak_mono_algebra less_eq less IE
proof (unfold_locales)
  fix f ls rs a b
  { fix as bs :: "'a list"
    assume "length as = length bs"
    with var_domain[of _ f]
    have "(\<forall>i < length as. as ! i \<sqsubseteq> bs ! i) \<Longrightarrow> IE f as \<sqsubseteq> IE f bs"
      by (auto simp: IE_def intro: target.mono_range)
  }
  note * = this
  assume ab: "a \<sqsubseteq> b"
  show "IE f (ls @ a # rs) \<sqsubseteq> IE f (ls @ b # rs)"
  proof(intro * allI impI, simp)
    fix i assume "i < length (ls @ a # rs)"
    from ab show "(ls @ a # rs) ! i \<sqsubseteq> (ls @ b # rs) ! i"
    by(cases i "length ls" rule: linorder_cases, auto simp: nth_append)
  qed
qed

end

locale encoded_wf_weak_mono_algebra =
  encoded_weak_mono_algebra less_eq less I E + wf_order less_eq less for less_eq less I E
begin

sublocale target: encoded_wf_algebra less_eq less I E..

sublocale wf_weak_mono_algebra less_eq less IE..

end

section {* Max-polynomial Interpretations as an Instance *}

text {* As a target algebra, we introduce a $G$-algebra where $G = \{+,\times,\max\}$. *}

locale maxpoly begin

datatype sig = Fconst nat | Fsum | Fprod | Fmax

fun I where
  "I (Fconst n) = (\<lambda>x. n)"
| "I Fsum = sum_list"
| "I Fprod = prod_list"
| "I Fmax = Max \<circ> set"

sublocale wf_weak_mono_algebra less_eq less I
proof
  fix f ls rs and a b :: nat
  assume ba: "b \<le> a"
  then show "I f (ls @ b # rs) \<le> I f (ls @ a # rs)"
  proof(cases f)
    case Fmax
    note ba
    also have "a \<le> Max (insert a (set ls \<union> set rs))" by auto
    finally show ?thesis by (auto simp: Fmax)
  qed auto
qed

end

text {* Now we encode arbitrary $F$-terms into the $G$-terms. *}

locale maxpoly_interpretation = encoding E
  for E :: "'f \<times> nat \<Rightarrow> (maxpoly.sig, nat) term"
begin

sublocale target: maxpoly.

sublocale encoded_wf_weak_mono_algebra "less_eq::nat\<Rightarrow>nat\<Rightarrow>bool" less maxpoly.I E
  by (unfold_locales, fact target.weak_mono)

text {* Now we get a reduction pair for free! *}
thm redpair.redpair_axioms

end


















end
